from typing import (
    Optional,
    cast,
)
from argparse import (
    ArgumentParser,
    Namespace
)
from datetime import datetime

from azure.storage.blob import ContainerClient

from .base_storage import BaseStorage


class BlobStorage(BaseStorage):
    _container_client: Optional[ContainerClient]
    max_retention_days: int

    def __init__(self):
        self._container_client = None

    def quilla_addopts(self, parser: ArgumentParser):
        '''
        Adds the appropriate CLI arguments

        Args:
            parser: The Quilla argument parser
        '''
        az_group = parser.add_argument_group(
            title='Azure Blob Storage Options'
        )

        az_group.add_argument(
            '--connection-string',
            dest='connection_string',
            action='store',
            default=None,
            help='A connection string for the azure storage account to be used'
        )

        az_group.add_argument(
            '--container-name',
            dest='container_name',
            action='store',
            default='quilla',
            help='The name of the container that will be used for storing Quilla images'
        )

        az_group.add_argument(
            '--retention-days',
            dest='retention_days',
            type=int,
            default=30,
            help='The maximum number of days that reports should be allowed to exist. '
            'Reports older than this amount of days will be deleted. Set to -1 to let '
            'reports be kept indefinitely.'
        )

    def configure(
        self,
        connection_string: str,
        container_name: str,
        retention_days: int,
    ):
        '''
        Configure the container client and other necessary data, such as the max cleanup time.

        If a container with that name does not exist, it will be created.

        Args:
            connection_string: The full connection string to the storage account
            container_name: The name of the container that should be used to store all images
            retention_days: The maximum number of days a report should be allowed to have before
                being cleaned up
        '''
        self._container_client = client = ContainerClient.from_connection_string(
            connection_string,
            container_name=container_name
        )

        self.max_retention_days = retention_days

        if not client.exists():
            client.create_container()

    @property
    def container_client(self) -> ContainerClient:
        '''
        An instance of the container client, casting it to ContainerClient.

        This should be used exclusively from the abstract methods from BaseStorage

        Returns:
            The container client
        '''

        return cast(ContainerClient, self._container_client)

    def quilla_configure(self, args: Namespace):
        '''
        Configures the plugin to run

        Args:
            args: A namespace generated by parsing the args from the CLI
        '''
        if args.connection_string is None:
            return

        self.configure(
            args.connection_string,
            args.container_name,
            args.retention_days,
        )

    @property
    def is_enabled(self) -> bool:
        return self._container_client is not None

    def find_image_by_baseline(self, baseline_id: str) -> bytes:
        blob = self.container_client.get_blob_client(f'baselines/{baseline_id}.png')

        if not blob.exists():
            return b''

        downloader = blob.download_blob()

        blob_data = downloader.readall()

        return blob_data

    def store_baseline_image(self, run_id: str, baseline_id: str, baseline: bytes) -> str:
        blob = self.container_client.get_blob_client(f'baselines/{baseline_id}.png')
        snapshot = self.container_client.get_blob_client(
            f'baselines/snapshots/{run_id}/{baseline_id}.png'
        )

        blob.upload_blob(baseline)
        snapshot.upload_blob(baseline)

        return blob.url

    def store_treatment_image(self, run_id: str, baseline_id: str, treatment: bytes) -> str:
        blob = self.container_client.get_blob_client(f'runs/{run_id}/{baseline_id}_treatment.png')
        blob.upload_blob(treatment)

        return blob.url

    def make_baseline_uri(self, run_id: str, baseline_id: str) -> str:
        baseline_data = self.find_image_by_baseline(baseline_id)
        blob = self.container_client.get_blob_client(f'runs/{run_id}/{baseline_id}.png')

        blob.upload_blob(baseline_data)

        return blob.url

    def cleanup_reports(self):
        blobs = self.container_client.list_blobs()
        current_time = datetime.now()

        for blob in filter(lambda x: 'runs' in x.name, blobs):
            time_created: datetime = datetime.fromtimestamp(blob.creation_time.timestamp())

            delta = current_time - time_created
            if self.max_retention_days > -1 and delta.days > self.max_retention_days:
                blob_client = self.container_client.get_blob_client(blob)
                blob_client.delete_blob()



from typing import (
    Dict,
    cast
)
from quilla.common.enums import (
    XPathValidationStates,
    ValidationTypes
)
from quilla.reports.validation_report import ValidationReport


class VisualParityReport(ValidationReport):
    '''
    Report type for VisualParity validation state

    Args:
        target: The validation target
        browser_name: The name of the browser that the validation was performed on
        success: Whether the validation passed or not
        msg: An optional string adding further context to the report
        baseline_id: The ID for the image baseline
        baseline_image_uri: A URI that allows locating the baseline image (i.e. path, link, etc)
        treatment_image_uri: A URI that allows locating the treatment image (i.e. path, link, etc)
        delta_image_uri: A URI that allows locating the delta image (i.e. path, link, etc)

    Attributes:
        validation_type: The string representation of the type of validation performed
        target: The validation target
        state: The desired state used for validation
        success: Whether the validation passed or not
        msg: An optional string adding further context to the report
        baseline_id: The ID for the image baseline
        baseline_image_uri: A URI that allows locating the baseline image (i.e. path, link, etc)
        treatment_image_uri: A URI that allows locating the treatment image (i.e. path, link, etc)
        delta_image_uri: A URI that allows locating the delta image (i.e. path, link, etc)
    '''

    def __init__(
        self,
        target: str,
        browser_name: str,
        success: bool,
        baseline_id: str,
        msg: str = '',
        baseline_image_uri: str = '',
        treatment_image_uri: str = '',
        delta_image_uri: str = '',
    ):
        super().__init__(
            validation_type=ValidationTypes.XPATH.value,
            target=target,
            state=XPathValidationStates.VISUAL_PARITY.value,
            browser_name=browser_name,
            success=success,
            msg=msg
        )
        self.baseline_id = baseline_id
        self.baseline_image_uri = baseline_image_uri
        self.treatment_image_uri = treatment_image_uri
        self.delta_image_uri = delta_image_uri

    def to_dict(self):
        report = super().to_dict()

        report_data = report['validationReport']

        report_data['baselineId'] = self.baseline_id

        if self.baseline_image_uri:
            report_data['baselineImageUri'] = self.baseline_image_uri

        if self.treatment_image_uri:
            report_data['treatmentImageUri'] = self.treatment_image_uri

        if self.delta_image_uri:
            report_data['deltaImageUri'] = self.delta_image_uri

        return {
            'visualParityReport': report_data
        }

    @classmethod
    def from_dict(cls, report) -> 'VisualParityReport':
        params: Dict[str, str] = report['visualParityReport']
        msg = params.get('msg', '')
        baseline_id = params['baselineId']
        baseline_uri = params.get('baselineImageUri', '')
        treatment_uri = params.get('treatmentImageUri', '')
        delta_uri = params.get('deltaImageUri', '')
        success = cast(bool, params['passed'])

        return VisualParityReport(
            target=params['target'],
            browser_name=params['targetBrowser'],
            success=success,
            msg=msg,
            baseline_id=baseline_id,
            baseline_image_uri=baseline_uri,
            treatment_image_uri=treatment_uri,
            delta_image_uri=delta_uri,
        )



import re
from typing import (
    Optional,
    Dict,
    Callable,
    List
)
from io import BytesIO

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.by import By
from PIL import Image

from quilla.ctx import Context
from quilla.common.enums import (
    XPathValidationStates,
    ValidationStates,
    ValidationTypes,
    VisualParityImageType,
)
from quilla.reports import (
    ValidationReport,
    VisualParityReport
)
from quilla.steps.base_steps import BaseValidation


class XPathValidation(BaseValidation):
    '''
    Class defining the behaviour for performing XPath validations

    Args:
        ctx: The runtime context for the application
        target: The XPath of the element to perform the validation against
        state: The desired state of the target web element
        driver: An optional argument to allow the driver to be bound at object creation.
    '''
    def __init__(
        self,
        ctx: Context,
        target: str,
        state: XPathValidationStates,
        parameters: Optional[Dict],
        driver: Optional[WebDriver] = None,
    ) -> None:
        selector: Dict[ValidationStates, Callable[[], ValidationReport]] = {
            XPathValidationStates.EXISTS: self._check_exists,
            XPathValidationStates.NOT_EXISTS: self._check_not_exists,
            XPathValidationStates.VISIBLE: self._check_visible,
            XPathValidationStates.NOT_VISIBLE: self._check_not_visible,
            XPathValidationStates.TEXT_MATCHES: self._check_text_matches,
            XPathValidationStates.NOT_TEXT_MATCHES: self._check_not_text_matches,
            XPathValidationStates.HAS_PROPERTY: self._check_has_property,
            XPathValidationStates.NOT_HAS_PROPERTY: self._check_not_has_property,
            XPathValidationStates.PROPERTY_HAS_VALUE: self._check_property_has_value,
            XPathValidationStates.NOT_PROPERTY_HAS_VALUE: self._check_not_property_has_value,
            XPathValidationStates.HAS_ATTRIBUTE: self._check_has_attribute,
            XPathValidationStates.NOT_HAS_ATTRIBUTE: self._check_not_has_attribute,
            XPathValidationStates.ATTRIBUTE_HAS_VALUE: self._check_attribute_has_value,
            XPathValidationStates.NOT_ATTRIBUTE_HAS_VALUE: self._check_not_attribute_has_value,
            XPathValidationStates.VISUAL_PARITY: self._check_visual_parity,
        }
        super().__init__(
            ctx,
            ValidationTypes.XPATH,
            target,
            state,
            selector,
            parameters=parameters,
            driver=driver
        )

    def _find_all(self) -> List[WebElement]:
        '''
        Proxy method to find all elements specified by the _target attribute

        Returns:
            A list of all the elements found for that specific target, searched by XPath

        Raises:
            NoDriverException: If the driver is not currently bound to this step
        '''
        return self.driver.find_elements(By.XPATH, self.target)

    def _element_text_matches_pattern(self) -> bool:
        self._verify_parameters('pattern')
        element_text = self.element.text
        pattern = self.parameters['pattern']

        return re.search(pattern, element_text) is not None

    def _element_exists(self) -> bool:
        return len(self._find_all()) > 0

    def _element_visible(self) -> bool:
        return self.element.is_displayed()

    def _element_has_property(self) -> bool:
        self._verify_parameters('name')

        return self.element.get_property(self.parameters['name']) is not None

    def _element_has_attribute(self) -> bool:
        self._verify_parameters('name')

        return self.element.get_attribute(self.parameters['name']) is not None

    def _element_check_value(self, value_fetch_fn: Callable[[str], Optional[str]]) -> bool:
        self._verify_parameters('name', 'value')

        element_value = value_fetch_fn(self.parameters['name'])

        return element_value == self.parameters['value']

    def _element_property_has_value(self) -> bool:
        return self._element_check_value(self.element.get_property)

    def _element_attribute_has_value(self) -> bool:
        return self._element_check_value(self.element.get_attribute)

    def _check_exists(self) -> ValidationReport:
        return self._create_report(
            self._element_exists()
        )

    def _check_not_exists(self) -> ValidationReport:
        return self._create_report(
            not self._element_exists
        )

    def _check_visible(self) -> ValidationReport:
        return self._create_report(
            self._element_visible()
        )

    def _check_not_visible(self) -> ValidationReport:
        return self._create_report(
            not self._element_visible
        )

    def _check_text_matches(self) -> ValidationReport:
        text_matches = self._element_text_matches_pattern()
        msg = ''

        if not text_matches:
            msg = (
                f'Element text "{self.element.text}" '
                f'does not match pattern "{self._parameters["pattern"]}"'
            )
        return self._create_report(
            text_matches,
            msg
        )

    def _check_not_text_matches(self) -> ValidationReport:
        text_matches = self._element_text_matches_pattern()
        msg = ''

        if text_matches:
            msg = (
                f'Element text "{self.element.text}" '
                f'matches pattern "{self._parameters["pattern"]}"'
            )
        return self._create_report(
            not text_matches,
            msg
        )

    def _check_has_property(self) -> ValidationReport:
        return self._create_report(
            self._element_has_property()
        )

    def _check_not_has_property(self) -> ValidationReport:
        return self._create_report(
            not self._element_has_property()
        )

    def _check_has_attribute(self) -> ValidationReport:
        return self._create_report(
            self._element_has_attribute()
        )

    def _check_not_has_attribute(self) -> ValidationReport:
        return self._create_report(
            not self._element_has_attribute()
        )

    def _check_property_has_value(self) -> ValidationReport:
        return self._create_report(
            self._element_property_has_value()
        )

    def _check_not_property_has_value(self) -> ValidationReport:
        return self._create_report(
            not self._element_property_has_value()
        )

    def _check_attribute_has_value(self) -> ValidationReport:
        return self._create_report(
            self._element_attribute_has_value()
        )

    def _check_not_attribute_has_value(self) -> ValidationReport:
        return self._create_report(
            not self._element_attribute_has_value
        )

    def _check_visual_parity(self) -> ValidationReport:
        self._verify_parameters('baselineID')

        baseline_id = self.parameters['baselineID']
        update_baseline = (
            self.ctx.update_all_baselines or
            baseline_id in self.ctx.update_baseline
        )

        if update_baseline:
            result = self.ctx.pm.hook.quilla_store_image(
                ctx=self.ctx,
                baseline_id=baseline_id,
                image_bytes=self.element.screenshot_as_png,
                image_type=VisualParityImageType.BASELINE
            )

            if result is None:
                return VisualParityReport(
                    success=False,
                    target=self._target,
                    browser_name=self.driver.name,
                    baseline_id=baseline_id,
                    msg='No baseline storage mechanism configured'
                )

            baseline_uri = result

            if baseline_uri == '':
                return VisualParityReport(
                    success=False,
                    target=self._target,
                    browser_name=self.driver.name,
                    baseline_id=baseline_id,
                    msg='Unable to update the baseline image'
                )

            return VisualParityReport(
                success=True,
                target=self._target,
                browser_name=self.driver.name,
                baseline_id=baseline_id,
                baseline_image_uri=baseline_uri,
                msg='Successfully updated baseline URI'
            )

        treatment_image_bytes = self.element.screenshot_as_png
        treatment_image = Image.open(BytesIO(treatment_image_bytes))

        plugin_result = self.ctx.pm.hook.quilla_get_visualparity_baseline(
            ctx=self.ctx,
            baseline_id=baseline_id
        )

        if plugin_result is None:
            return VisualParityReport(
                success=False,
                target=self._target,
                browser_name=self.driver.name,
                baseline_id=baseline_id,
                msg='No baseline storage mechanism configured'
            )

        if len(plugin_result) == 0:
            if self.ctx.create_baseline_if_none:
                baseline_uri = self.ctx.pm.hook.quilla_store_image(
                    ctx=self.ctx,
                    baseline_id=baseline_id,
                    image_bytes=self.element.screenshot_as_png,
                    image_type=VisualParityImageType.BASELINE
                )
                return VisualParityReport(
                    success=True,
                    target=self._target,
                    browser_name=self.driver.name,
                    baseline_id=baseline_id,
                    baseline_image_uri=baseline_uri,
                    msg='Successfully updated baseline URI'
                )

            return VisualParityReport(
                success=False,
                target=self._target,
                browser_name=self.driver.name,
                baseline_id=baseline_id,
                msg='No baseline image found'
            )

        baseline_image_bytes = plugin_result
        baseline_image = Image.open(BytesIO(baseline_image_bytes))

        success = baseline_image == treatment_image  # Run the comparison with Pillow

        if success:
            return VisualParityReport(
                success=success,
                target=self._target,
                browser_name=self.driver.name,
                baseline_id=baseline_id,
            )

        treatment_uri = self.ctx.pm.hook.quilla_store_image(
            ctx=self.ctx,
            baseline_id=baseline_id,
            image_bytes=treatment_image_bytes,
            image_type=VisualParityImageType.TREATMENT,
        )

        baseline_uri = self.ctx.pm.hook.quilla_get_baseline_uri(
            ctx=self.ctx,
            run_id=self.ctx.run_id,
            baseline_id=baseline_id
        )

        return VisualParityReport(
            success=False,
            target=self._target,
            browser_name=self.driver.name,
            baseline_id=baseline_id,
            baseline_image_uri=baseline_uri,
            treatment_image_uri=treatment_uri,
        )



'''
Entrypoint code for the Quilla module. Deals with setting up the parser and
the runtime context for the application, then executing the rest of the application
'''
import argparse
import sys
import json
from typing import (
    List,
)
import logging
from pkg_resources import get_distribution

from quilla.ui_validation import QuillaTest
from quilla.ctx import (
    Context,
    get_default_context
)
from quilla.reports import ReportSummary
from quilla.plugins import get_plugin_manager


def make_parser() -> argparse.ArgumentParser:  # pragma: no cover
    '''
    Creates the required parser to run Quilla from the command line


    Returns:
        A pre-configured ArgParser instance
    '''
    parser = argparse.ArgumentParser(
        prog='quilla',
        description='''
        Program to provide a report of UI validations given a json representation
        of the validations or given the filename containing a json document describing
        the validations
        ''',
    )

    parser.add_argument(
        '--version',
        action='store_true',
        help='Prints the version of the software and quits'
    )

    data_group = parser.add_mutually_exclusive_group()

    data_group.add_argument(
        '-f',
        '--file',
        dest='file_name',
        action='store',
        help='A file containing a Quilla test',
        default=None,
    )
    data_group.add_argument(
        '-r',
        '--raw',
        action='store',
        help='A Quilla test passed in as a raw string',
        default=None,
    )

    config_group = parser.add_argument_group(title='Configuration options')
    config_group.add_argument(
        '-i',
        '--run-id',
        action='store',
        metavar='run_id',
        default=None,
        help='A run ID for quilla, if manually passed in. '
        'Used to set many quilla tests to have the same run ID '
        'If no run ID is manually passed in, it will be auto-generated'
    )
    config_group.add_argument(
        '-u',
        '--update-baseline',
        dest='update_baseline',
        action='append',
        metavar='BASELINE_ID',
        help='Used to update the specific baseline image associated with '
        'the provided baseline ID for VisualParity. '
        'Different plugins define different behaviour, so this is not '
        'necessarily a lossless operation.',
        default=[],
    )
    config_group.add_argument(
        '-U',
        '--update-all-baselines',
        dest='update_all_baselines',
        action='store_true',
        help='Used to update all baseline images in a Quilla test file. '
        'this is equivalent to using --update-baseline with every baseline ID '
        'that is used in the Quilla test.'
    )
    config_group.add_argument(
        '--create-baseline-if-absent',
        dest='create_baseline_if_none',
        action='store_true',
        help='A flag to set that will create a new baseline image if a storage '
        'mechanism is configured but no baseline image is found with the specified ID.'
    )
    config_group.add_argument(
        '-d',
        '--definitions',
        action='append',
        metavar='file',
        help='A file with definitions for the \'Definitions\' context object'
    )
    config_group.add_argument(
        '--driver-dir',
        dest='drivers_path',
        action='store',
        default='.',
        help='The directory where browser drivers are stored',
    )
    config_group.add_argument(
        '--no-sandbox',
        dest='no_sandbox',
        action='store_true',
        help='''
        Adds \'--no-sandbox\' to the Chrome and Edge browsers.
        Useful for running in docker containers'
        '''
    )

    output_group = parser.add_argument_group(title='Output Options')
    output_group.add_argument(
        '-P',
        '--pretty',
        action='store_true',
        help='Set this flag to have the output be pretty-printed'
    )
    output_group.add_argument(
        '--indent',
        type=int,
        default=4,
        help='How much space each indent level should have when pretty-printing the report'
    )

    debug_group = parser.add_argument_group(title='Debug Options')
    debug_group.add_argument(
        '--debug',
        action='store_true',
        help='Enable debug mode',
    )
    debug_group.add_argument(
        '-v',
        '--verbose',
        action='count',
        help='Flag to increase the verbosity of the outputs. '
        'Log outputs are directed to stderr by default.',
        default=0
    )

    # Sets the application handler, i.e. the function that runs when ctx.run() is called
    parser.set_defaults(handler=run)

    return parser


def get_early_configs(parser: argparse.ArgumentParser, args: List[str]) -> argparse.Namespace:
    '''
    Extracts known configs early on. This will disable help/usage printing and prevent
    exiting on failure, since it potentially needs to ignore errors to collect the early configs

    Args:
        parser: A parser instance, such as the one returned by `make parser`
        args: A list of strings containing the arguments to be consumed

    Returns:
        A namespace object containing the early configurations for Quilla

    '''

    # Saves functions that need to be disabled for early configs
    exit_function = parser.exit
    print_help_fn = parser.print_help
    print_usage_fn = parser.print_usage

    # "Disable" the above functions
    do_nothing = lambda *x: None  # noqa: E731
    parser.exit = do_nothing  # type: ignore
    parser.print_help = do_nothing  # type: ignore
    parser.print_usage = do_nothing  # type: ignore

    # Get early args
    parsed_args, _ = parser.parse_known_intermixed_args(args)

    # Revert the disabled functions to their normal usage
    parser.exit = exit_function  # type: ignore
    parser.print_help = print_help_fn  # type: ignore
    parser.print_usage = print_usage_fn  # type: ignore

    return parsed_args


def make_default_logger(early_configs: argparse.Namespace) -> logging.Logger:
    '''
    Creates a default logger class for Quilla. All internal logger setup logic
    is done in this function

    Args:
        early_configs: The extracted configurations from an early pass-through
            of the parse_known_intermixed_args function, before plugins run

    Returns:
        A configured ``Logger``
    '''
    logger = logging.getLogger('quilla')

    logger.propagate = False  # Disable the root logger

    formatter = logging.Formatter(
        fmt='[%(asctime)s] [%(levelname)s] - %(message)s',
        datefmt='%H:%M:%S'
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    logger.addHandler(stream_handler)

    if early_configs.verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif early_configs.verbose == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    if early_configs.debug:
        logger.setLevel(logging.DEBUG)

    return logger


def execute(ctx: Context) -> ReportSummary:
    '''
    Runs all defined UI validations from the json, either from file or from raw json text, and
    prints all the resulting reports to the console

    Args:
        json: The json string describing the validation

    Returns:
        A summary of all reports produced by Quilla
    '''

    ctx.logger.debug('Building UIValidation object from JSON')
    quilla_test = QuillaTest.from_json(ctx, ctx.json)

    ctx.logger.info('Passing UIValidation instance to "quilla_prevalidate" hook')
    ctx.pm.hook.quilla_prevalidate(validation=quilla_test)

    ctx.logger.debug('Running all validations')
    reports = quilla_test.validate_all()

    ctx.logger.info('Running "quilla_postvalidate" hooks')
    ctx.pm.hook.quilla_postvalidate(ctx=ctx, reports=reports)

    return reports


def setup_context(
    args: List[str],
    plugin_root: str = '.',
    recreate_context: bool = False,
) -> Context:
    '''
    Starts up the plugin manager, creates parser, parses args and sets up the application context

    Args:
        args: A list of cli options, such as sys.argv[1:]
        plugin_root: The directory used by the plugin manager to search for `uiconf.py` files
        recreate_context: Whether the context should be recreated
    Returns:
        A runtime context configured by the hooks and the args
    '''
    parser = make_parser()
    early_configs = get_early_configs(parser, args)

    if early_configs.version:
        print(f'Quilla v{get_distribution("quilla").version}')
        sys.exit(0)

    logger = make_default_logger(early_configs)

    logger.debug('Default logger configured, running Quilla with args %s', args)

    logger.debug('Initializing plugin manager')

    pm = get_plugin_manager(plugin_root, logger)

    logger.info('Running "quilla_configure_logger" hook')
    pm.hook.quilla_configure_logger(logger=logger)

    logger.info('Running "quilla_addopts" hook')
    pm.hook.quilla_addopts(parser=parser)  # type: ignore

    logging.debug('Parsing all args')
    parsed_args = parser.parse_args(args)

    # Set to empty list since argparse defaults to None
    if not parsed_args.definitions:
        parsed_args.definitions = []

    if parsed_args.file_name is None:
        json_data = parsed_args.raw
    else:
        with open(parsed_args.file_name) as f:
            json_data = f.read()

    logger.debug('Initializing context object')

    ctx = get_default_context(
        pm,
        parsed_args.debug,
        parsed_args.drivers_path,
        parsed_args.pretty,
        json_data,
        parsed_args.file_name is not None,
        parsed_args.no_sandbox,
        parsed_args.definitions,
        logger=logger,
        run_id=parsed_args.run_id,
        indent=parsed_args.indent,
        update_all_baselines=parsed_args.update_all_baselines,
        update_baseline=parsed_args.update_baseline,
        create_baseline_if_none=parsed_args.create_baseline_if_none,
        args=parsed_args,
        recreate_context=recreate_context,
    )

    logger.info('Running "quilla_configure" hook')
    pm.hook.quilla_configure(ctx=ctx, args=parsed_args)

    return ctx


def run(ctx: Context):
    '''
    Runs all reports and prints to stdout while providing the proper
    exit code

    Args:
        ctx: The application context
    '''
    reports = execute(ctx)

    ctx.logger.debug('Finished generating reports')

    out = reports.to_dict()

    if ctx.pretty:
        print(json.dumps(
            out,
            indent=ctx.pretty_print_indent,
            sort_keys=True
        ))
    else:
        print(json.dumps(out))

    if reports.fails > 0:
        exit_code = 1
    else:
        exit_code = 0

    sys.exit(exit_code)


def main():
    '''
    Creates the context and parses all arguments, then runs the default handler function
    '''
    ctx = setup_context(sys.argv[1:])
    ctx.logger.debug('Context setup complete, running the Quilla handler')

    ctx.run()


if __name__ == '__main__':
    main()



'''
A plugin to add LocalStorage functionality for the VisualParity plugin.
'''


from argparse import (
    ArgumentParser,
    Namespace,
)
from pathlib import Path
from typing import (
    Optional,
    cast
)

from .base_storage import BaseStorage


class LocalStorage(BaseStorage):
    baseline_directory: Optional[Path]
    runs_directory: Optional[Path]

    def __init__(
        self,
        storage_directory: Optional[str] = None
    ):
        if storage_directory is None:
            self.baseline_directory = None
            self.runs_directory = None
            return

        self.configure(storage_directory)

    def configure(self, storage_directory: str):
        '''
        Initialize all the required data
        '''
        baseline_path = Path(storage_directory)
        self.baseline_directory = baseline_path / 'baselines'
        self.runs_directory = baseline_path / 'runs'
        self.runs_directory.mkdir(exist_ok=True)
        self.baseline_directory.mkdir(exist_ok=True)

    @property
    def is_enabled(self) -> bool:
        return self.baseline_directory is not None

    def run_path(self, run_id: str) -> Path:
        path = cast(Path, self.runs_directory) / run_id

        path.mkdir(exist_ok=True)

        return path

    def store_baseline_image(self, run_id: str, baseline_id: str, baseline: bytes) -> str:
        baseline_path = cast(Path, self.baseline_directory) / f'{baseline_id}.png'

        snapshot_path = baseline_path.parent / 'snapshots' / f'{baseline_id}_{run_id}.png'

        snapshot_path.parent.mkdir(exist_ok=True)

        baseline_path.touch()

        baseline_path.write_bytes(baseline)

        snapshot_path.touch()
        snapshot_path.write_bytes(baseline)

        return baseline_path.absolute().as_uri()

    def store_treatment_image(
        self,
        run_id: str,
        baseline_id: str,
        treatment: bytes
    ) -> str:

        run_path = self.run_path(run_id)

        image_path = run_path / f'{baseline_id}_treatment.png'

        image_path.write_bytes(treatment)

        return image_path.absolute().as_uri()

    def find_image_by_baseline(self, baseline_id: str) -> bytes:
        image_path = cast(Path, self.baseline_directory) / f'{baseline_id}.png'

        if not image_path.exists():
            return b''

        return image_path.read_bytes()

    def make_baseline_uri(self, run_id: str, baseline_id: str) -> str:
        image_data = self.find_image_by_baseline(baseline_id)

        run_path = self.run_path(run_id)

        image_path = run_path / f'{baseline_id}.png'

        image_path.touch()

        image_path.write_bytes(image_data)

        return image_path.absolute().as_uri()

    def cleanup_reports(self):
        '''
        Method left blank, as no cleanup is provided for LocalStorage since
        users are expected to have granular control over their own filesystems
        '''

    def quilla_addopts(self, parser: ArgumentParser):
        '''
        Using the Quilla hook to add a new group of CLI args to the parser
        '''

        ls_group = parser.add_argument_group(title='Local Storage Options')

        ls_group.add_argument(
            '--image-directory',
            dest='image_dir',
            action='store',
            default=None,
            help='The directory that should be used for the LocalStorage '
            'plugin to store VisualParity images'
        )

    def quilla_configure(self, args: Namespace):
        if args.image_dir is not None:
            self.configure(args.image_dir)



'''
Hooks that allow extension of the behaviour of steps, including output and validation
related steps. This also includes any hooks related to context expressions
'''
from typing import (
    Type,
    Dict,
    Tuple,
    Optional
)

from quilla.hookspecs._hookspec import hookspec
from quilla.ctx import Context
from quilla.common.enums import (
    UITestActions,
    VisualParityImageType,
)
from quilla.steps.base_steps import BaseStepFactory


StepFactorySelector = Dict[UITestActions, Type[BaseStepFactory]]


@hookspec
def quilla_step_factory_selector(selector: StepFactorySelector):
    '''
    A hook called immediately before resolving the step factory for a given step definition.
    This is used to register new step factories for custom step objects.

    Most custom steps should just add themselves to one of the non-factory selector hooks, but if
    a custom step requires complex logic it might be beneficial to register a factory to
    have more fine-grained control over the logic

    Args:
        selector: The factory selector dictionary.
    '''


@hookspec(firstresult=True)
def quilla_context_obj(ctx: Context, root: str, path: Tuple[str]) -> Optional[str]:
    '''
    A hook to allow pluggins to resolve a context object given its root
    and a path. All plugins that implement this hook *must* return None if they cannot
    resolve the context object.

    It is not possible to override the default context object handlers

    Args:
        ctx: The runtime context for the application
        root: The name of the context object, which is expressed as the root
            of a dot-separated path in the validation files
        path: The remainder of the context object path, where data is being
            retrieved from

    Returns
        the data stored at the context object if existing, None otherwise
    '''


@hookspec(firstresult=True)
def quilla_store_image(
    ctx: Context,
    baseline_id: str,
    image_bytes: bytes,
    image_type: VisualParityImageType,
) -> Optional[str]:
    '''
    A hook to allow pluggins to store images for the VisualParity validation.

    Plugins should first check if they are able to authenticate or configure
    with whatever storage mechanism is being used to keep the baseline images,
    and return None if they cannot both read and write files. It is assumed that
    only one storage mechanism will be used for baseline/treatment images.

    If a plugin can write files to its storage mechanism, it should return the URI
    (i.e. the link, path, etc. that can be used to find the new image) so that it
    can be included in the report.

    Plugins to add storage mechanisms to VisualParity validations should implement
    this hook as well as the ``quilla_get_visualparity_baseline`` hook and some
    configuration method (i.e. adding CLI options with ``quilla_addopts`` or
    pulling data from the environment)

    Args:
        ctx: The runtime context for the application
        baseline_id: The image baseline ID associated with the image
        image_bytes: The data for the image PNG in bytes form
        image_type: The kind of image that it is, since different image types
            might be desired to be stored differently

    Returns
        An identifier that can locate the new image
    '''


@hookspec(firstresult=True)
def quilla_get_baseline_uri(ctx: Context, run_id: str, baseline_id: str) -> Optional[str]:
    '''
    A hook to allow plugins to retrieve some URI for the image associated with
    the baseline ID.

    Args:
        ctx: The runtime context of the application
        run_id: The run ID for the current run, in case the plugin tracks
            baselines for each run
        baseline_id: The unique ID for the baseline image

    Returns:
        An identifier that can locate the baseline image
    '''


@hookspec(firstresult=True)
def quilla_get_visualparity_baseline(ctx: Context, baseline_id: str) -> Optional[bytes]:
    '''
    A hook to allow pluggins to find baseline images for the VisualParity
    validation, called while the VisualParity validation step is being executed.

    Plugins should first check if they are able to authenticate or configure
    with whatever storage mechanism is being used to keep the baseline images,
    and return None if they cannot both read and write files. It is assumed that
    only one storage mechanism will be used for baseline/treatment images.

    If a plugin can read the files from its storage mechanism, it should search for the
    image by its baseline ID and return the data in bytes.

    Args:
        ctx: The runtime context for the application
        baseline_id: The baseline ID to search for

    Returns:
        The image data in bytes
    '''



'''
This module contains a base class that can be useful when defining new
storage plugins, as there are some behaviours that will be shared among
the configuration objects
'''

from abc import (
    ABC,
    abstractproperty,
    abstractmethod
)
from typing import (
    Optional
)

from quilla.ctx import Context
from quilla.common.enums import VisualParityImageType


class BaseStorage(ABC):
    @abstractproperty
    def is_enabled(self) -> bool:
        '''
        A property to be used to determine if the plugin is configured and
        therefore able to run

        Returns:
            True if the plugin is configured, False otherwise
        '''

    @abstractmethod
    def find_image_by_baseline(self, baseline_id: str) -> bytes:
        '''
        Searches the defined storage method for the image matching some
        baseline ID

        Args:
            baseline_id: The unique ID to search for. It is assumed every
                baseline ID is unique regardless of what test is requesting it.

        Returns:
            The bytes representation of the stored image if found, or an empty
            bytes object if the image is not found.
        '''

    @abstractmethod
    def cleanup_reports(self):
        '''
        Searches for reports that match some cleanup criteria, and deletes them
        if necessary. Not every storage plugin will implement logic for this function,
        choosing instead to have all images exist indefinitely.
        '''

    @abstractmethod
    def store_treatment_image(
        self,
        run_id: str,
        baseline_id: str,
        treatment: bytes,
    ) -> str:
        '''
        Stores a treatment image within the storage mechanism enabled by the plugin

        Args:
            run_id: The run ID of the current Quilla run, to version the treatment images
            baseline_id: The ID of the baseline that this treatment image is associated
                with
            treatment: The image data in bytes

        Returns:
            An identifier that can locate the newly stored treatment image
        '''

    @abstractmethod
    def store_baseline_image(
        self,
        run_id: str,
        baseline_id: str,
        baseline: bytes,
    ) -> str:
        '''
        Stores a baseline image under the given baseline_id.

        This function should be used to update the current baseline
        image, or create a new one if the baseline did not previously exist.

        The run ID is passed in case a plugin would like to use it to version and
        store previous image baselines

        Args:
            run_id: The ID of the current run of Quilla
            baseline_id: A unique identifier for the image
            baseline: The image data in bytes

        Returns:
            A URI for the new baseline image
        '''

    @abstractmethod
    def make_baseline_uri(
        self,
        run_id: str,
        baseline_id: str
    ) -> str:
        '''
        Generates a baseline URI for the current run given the baseline_id of the image.

        It is recommended that plugins create a clone of the baseline image
        when generating the URI so that the returned URI will uniquely identify
        the baseline that was used for the associated run ID. This ensures that
        even if the baseline image is updated, the report is still meaningful.

        Args:
            run_id: The unique ID identifying the current run
            baseline_id: The unique identifier for the image

        Returns:
            A URI that can locate the baseline image used for the given run
        '''

    def get_image(self, baseline_id: str) -> Optional[bytes]:
        '''
        Determines if the plugin should run, and if so searches for the image
        with the specified baseline ID and returns the byte data for it

        Args:
            baseline_id: The unique ID to search for

        Returns:
            None if the plugin is not enabled, a ``bytes`` representation
            of the image if it is. If no baseline image is found, returns an
            empty byte string
        '''
        if not self.is_enabled:
            return None

        return self.find_image_by_baseline(baseline_id)

    def get_baseline_uri(self, run_id: str, baseline_id: str) -> Optional[str]:
        '''
        Retrieves the URI for the baseline image

        Args:
            run_id: The unique ID for the current run of Quilla
            baseline_id: The unique ID for the baseline image

        Returns:
            None if the plugin is not enabled, a string URI if it is
        '''
        if not self.is_enabled:
            return None

        return self.make_baseline_uri(run_id, baseline_id)

    def quilla_store_image(
        self,
        ctx: Context,
        baseline_id: str,
        image_bytes: bytes,
        image_type: VisualParityImageType,
    ) -> Optional[str]:
        '''
        Stores a given image based on its type and possibly the run ID

        Args:
            ctx: The runtime context for Quilla
            baseline_id: The unique identifier for the image
            image_bytes: The byte data for the image
            image_type: The kind of image that is being stored

        Returns:
            A URI for the image that was stored, or None if the plugin
            is not enabled. The URI might be the empty string if the
            image type is not supported
        '''
        if not self.is_enabled:
            return None

        run_id = ctx.run_id

        image_uri = ''

        function_selector = {
            VisualParityImageType.TREATMENT: self.store_treatment_image,
            VisualParityImageType.BASELINE: self.store_baseline_image,
        }

        store_image_function = function_selector[image_type]

        image_uri = store_image_function(
            run_id,
            baseline_id,
            image_bytes
        )

        self.cleanup_reports()

        return image_uri

    def quilla_get_baseline_uri(self, run_id: str, baseline_id: str) -> Optional[str]:
        return self.get_baseline_uri(run_id, baseline_id)

    def quilla_get_visualparity_baseline(self, baseline_id: str) -> Optional[bytes]:
        return self.get_image(baseline_id)



# flake8: noqa

from quilla.hookspecs._hookspec import hookspec

from .configuration import (
    quilla_addopts,
    quilla_configure,
    quilla_configure_logger,
    quilla_prevalidate,
    quilla_resolve_enum_from_name
)
from .reports import (
    quilla_postvalidate,
)
from .steps import (
    quilla_context_obj,
    quilla_get_baseline_uri,
    quilla_get_visualparity_baseline,
    quilla_step_factory_selector,
    quilla_store_image
)



from typing import (
    cast,
    Optional,
    Dict,
    Union,
    Type
)

from selenium.webdriver.remote.webdriver import WebDriver

from quilla.ctx import Context
from quilla.common.enums import (
    ValidationTypes,
    ValidationStates,
)
from quilla.steps.base_steps import (
    BaseStepFactory,
    BaseValidation,
)
from quilla.steps.validations.xpath import XPathValidation
from quilla.steps.validations.url import URLValidation


ValidationDictionary = Dict[str, Union[str, ValidationStates, ValidationTypes]]


class Validation(BaseStepFactory):
    '''
    Factory class for the different validations
    '''
    validation_selector: Dict[ValidationTypes, Type[BaseValidation]] = {
        ValidationTypes.XPATH: XPathValidation,
        ValidationTypes.URL: URLValidation,
    }

    @classmethod
    def from_dict(
        cls,
        ctx: Context,
        validation_dict: ValidationDictionary,
        driver: Optional[WebDriver] = None
    ) -> BaseValidation:
        '''
        From a validation dict, produces the appropriate validation object

        Args:
            ctx: The runtime context for the application
            validation_dict: A dictionary containing the definition of a validation, including
                the target, state, and type of validation to be performed
            driver: The driver that will be connected to the validation, if any

        Returns:
            Validation object of the type requested in the validation dictionary
        '''
        validation_params = {
            'driver': driver,
            'target': validation_dict['target'],
            'state': validation_dict['state'],
            'parameters': validation_dict.get('parameters', None),
        }

        validation_type = cast(ValidationTypes, validation_dict['type'])

        validation = cls.validation_selector[validation_type]

        return validation(ctx=ctx, **validation_params)  # type: ignore



'''
Module with all requried enums for the UIValidation
'''

from enum import Enum


# Enums
class ValidationTypes(Enum):
    '''
    The currently-supported types of validation allowed
    '''
    XPATH = 'XPath'
    URL = 'URL'


class ValidationStates(Enum):
    '''
    Base class for validation state enums
    '''


class XPathValidationStates(ValidationStates):
    '''
    States that the XPath validation class recognizes
    '''
    EXISTS = 'Exists'
    NOT_EXISTS = 'NotExists'
    VISIBLE = 'Visible'
    NOT_VISIBLE = 'NotVisible'
    TEXT_MATCHES = 'TextMatches'
    NOT_TEXT_MATCHES = 'NotTextMatches'
    HAS_PROPERTY = 'HasProperty'
    NOT_HAS_PROPERTY = 'NotHasProperty'
    PROPERTY_HAS_VALUE = 'PropertyHasValue'
    NOT_PROPERTY_HAS_VALUE = 'NotPropertyHasValue'
    HAS_ATTRIBUTE = 'HasAttribute'
    NOT_HAS_ATTRIBUTE = 'NotHasAttribute'
    ATTRIBUTE_HAS_VALUE = 'AttributeHasValue'
    NOT_ATTRIBUTE_HAS_VALUE = 'NotAttributeHasValue'
    VISUAL_PARITY = 'VisualParity'


class URLValidationStates(ValidationStates):
    '''
    States that the URL validation class recognizes
    '''
    CONTAINS = 'Contains'
    NOT_CONTAINS = 'NotContains'
    EQUALS = 'Equals'
    NOT_EQUALS = 'NotEquals'
    MATCHES = 'Matches'
    NOT_MATCHES = 'NotMatches'


class UITestActions(Enum):
    '''
    All supported UI test actions
    '''
    CLICK = 'Click'
    CLEAR = 'Clear'
    SEND_KEYS = 'SendKeys'
    WAIT_FOR_EXISTENCE = 'WaitForExistence'
    WAIT_FOR_VISIBILITY = 'WaitForVisibility'
    NAVIGATE_TO = 'NavigateTo'
    VALIDATE = 'Validate'
    REFRESH = 'Refresh'
    ADD_COOKIES = 'AddCookies'
    SET_COOKIES = 'SetCookies'
    REMOVE_COOKIE = 'RemoveCookie'
    CLEAR_COOKIES = 'ClearCookies'
    NAVIGATE_FORWARD = 'NavigateForward'
    NAVIGATE_BACK = 'NavigateBack'
    SET_BROWSER_SIZE = 'SetBrowserSize'
    HOVER = 'Hover'
    OUTPUT_VALUE = 'OutputValue'


class ReportType(Enum):
    '''
    All the currently supported report types
    '''
    VALIDATION = 'Validation'
    STEP_FAILURE = 'StepFailure'


class BrowserTargets(Enum):
    '''
    All the currently supported browser targets
    '''
    FIREFOX = 'Firefox'
    CHROME = 'Chrome'
    EDGE = 'Edge'


class OutputSources(Enum):
    '''
    Supported sources for the OutputValue action
    '''
    LITERAL = 'Literal'
    XPATH_TEXT = 'XPathText'
    XPATH_PROPERTY = 'XPathProperty'


class VisualParityImageType(Enum):
    '''
    The types of images that are supported by the
    VisualParity validation state
    '''

    BASELINE = 'Baseline'
    TREATMENT = 'Treatment'



from typing import (
    Optional,
    Dict,
    Callable,
)

from selenium.webdriver.remote.webdriver import WebDriver

from quilla.ctx import Context
from quilla.common.enums import (
    ValidationStates,
    ValidationTypes,
    URLValidationStates,
)
from quilla.reports import ValidationReport
from quilla.steps.base_steps import BaseValidation


class URLValidation(BaseValidation):
    '''
    Class defining the behaviour for performing URL validations

    Args:
        ctx: The runtime context for the application
        target: The URL to perform the validation with
        state: The desired state of the target url
        driver: An optional argument to allow the driver to be bound at object creation.
    '''
    def __init__(
        self,
        ctx: Context,
        target: str,
        state: URLValidationStates,
        parameters: Optional[Dict],
        driver: Optional[WebDriver] = None,
    ) -> None:
        selector: Dict[ValidationStates, Callable[[], ValidationReport]] = {
            URLValidationStates.CONTAINS: self._check_contains,
            URLValidationStates.NOT_CONTAINS: self._check_not_contains,
            URLValidationStates.EQUALS: self._check_equals,
            URLValidationStates.NOT_EQUALS: self._check_not_equals
        }
        super().__init__(
            ctx,
            ValidationTypes.URL,
            target,
            state,
            selector=selector,
            parameters=parameters,
            driver=driver
        )

    def perform(self) -> ValidationReport:
        '''
        Performs the correct action based on what is defined within the selector, and returns
        the resulting report produced.

        Returns:
            A report summarizing the results of the executed validation
        '''
        report = super().perform()

        if not report.success:
            report.msg = f'Expected URL: "{self._target}", Received URL "{self.url}"'

        return report

    @property
    def url(self) -> str:
        '''
        The current URL of the browser

        Raises:
            NoDriverException: If the driver is not currently bound to this step
        '''
        return self.driver.current_url

    def _check_contains(self) -> ValidationReport:
        return self._create_report(
            self.url.find(self.target) > -1
        )

    def _check_not_contains(self) -> ValidationReport:
        return self._create_report(
            self.url.find(self.target) == -1
        )

    def _check_equals(self) -> ValidationReport:
        return self._create_report(
            self.url == self.target
        )

    def _check_not_equals(self) -> ValidationReport:
        return self._create_report(
            self.url != self.target
        )
